/*

	graphics.c											9/30/2008


		8-bit ATARI computer graphics library using ANTIC mode E graphics.   
		
		Most of the library is written in the 6502 macro assembler used by the CC65 compiler, but
	this is the C interface to the assembler routines.
	
		The plot routine which is written in the assembler code and interfaces to C through two
	zero page plot coordinate variables and the corresponding rasterization functions:

		Zero Page variables:
			X0,Y0				: screen x,y coordinate 0
			X1,Y1				: screen x,y coordinate 1

		Plot functions:
			xy_plot_0()				: plot x,y coordinate 0
			xy_plot_1()				: plot x,y coordinate 1
		
		This two variable scheme is used in order to facilitate symmetric drawing functions like the 
	line drawing algorithm, which is the symmetric version of Bresenham's line drawing algorithm.
	
		Line function:
			line()					: draw line between coordinates 0 and 1
			
		There are two special line functions:
		
			h_line()				: draw horizontal line between coordinates 0 and 1
			v_line()				: draw vertical line between coordinates 0 and 1

		The more interesting of the two is h_line() function which draws polygon scanlines in the polygon
	rasterization code:
	
			fill_polygon()			: fill polygon with current color
		
		This code uses the same Bresenham's algorithm to determine the edges of the polygon, and then
	uses the horizontal line routine to fill in the polygon's scanlines.
	
		There are special cases for polygons, such as the rectangle, which turns out to be
	just an extended case of the h_line() routine in the assembler code.
	
			fill_rect()				: fill rectangle with current color
	
		The rest of the functions are coded in C, with the exception of the screen buffer switching 
	functions.
		
			init_graphics()			: initialize double buffered graphics (ANTIC mode E)
			set_color()				: set current drawing color
			clear_screen()			: fill entire screen with current color
			switch_screen()			: swap screen buffers
			exit_graphics()			: restore previous graphics mode

		author:

			Bart Jaszcz

		
*/

//#include <stdio.h>

#include "global.h"
#include "graphics.h"

// pixel masks which must remain in this order and cannot cross page boundries
byte pixel_mask[4] = { 0xc0, 0x30, 0x0c, 0x03 };	// bit mask shifted to each pixel within byte
byte left_mask[4]  = { 0xff, 0x3f, 0x0f, 0x03 };	// bit mask for left shift
byte right_mask[4] = { 0xff, 0xfc, 0xf0, 0xc0 };	// bit mask for right shift

// bytes filled with color values
byte pixel_fill[4] = { 0x00, 0x55, 0xaa, 0xff };

// individual screen buffers (foreground and background)
byte screen_count;             // number of screens allocated
byte current_screen;				// screen selector
byte *display[2];				// pointer to display list for each screen
byte *screen[2];				// pointer to each screen memory

// saved ANTIC registers
byte color_val[4];
byte sdmctl_val;
uint sdlstl_val;

// switch foreground/background screens signal for screen swapper
byte request_swap;

// scanline tables for each screen buffer
byte scan_low[2][HEIGHT];		// scanline addresses (low bytes)
byte scan_high[2][HEIGHT];	// scanline addresses (high bytes)

// number of bytes used in buffer
long buffer_usage;

// polygon rasterization data
byte left_edge[HEIGHT];		// polygon left edge
byte right_edge[HEIGHT];		// polygon right edge

// 4x6 font (GIMP export to XBM file)

#define FONT_CHARS 96
#define FONT_FIRST_CHAR 32
#define  FONT_WIDTH 4
#define  FONT_HEIGHT 6
#define FONT_BYTES (FONT_CHARS * FONT_HEIGHT / 2)

byte font4x6[FONT_BYTES] = {
   0x40, 0xaa, 0x24, 0x64, 0x24, 0x0a, 0x00, 0x80, 0x4c, 0x66, 0xe8, 0xe4,
   0x44, 0x00, 0x08, 0x62, 0x4e, 0xc6, 0xe6, 0xce, 0xea, 0xa8, 0xa2, 0x4a,
   0x46, 0xc6, 0xae, 0xaa, 0xaa, 0x6e, 0x62, 0x04, 0x06, 0x02, 0x08, 0x08,
   0x42, 0x24, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xc0, 0x64, 0x0a,
   0x40, 0xea, 0x8c, 0x2a, 0x42, 0x44, 0x00, 0x80, 0x6a, 0x88, 0x2a, 0x82,
   0xaa, 0x00, 0x04, 0x84, 0xaa, 0x2a, 0x2a, 0x22, 0x4a, 0xa8, 0xe2, 0xae,
   0xaa, 0x2a, 0xa4, 0xaa, 0xaa, 0x28, 0x42, 0x0a, 0x04, 0x02, 0x08, 0x04,
   0x02, 0x20, 0x04, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x40, 0x44, 0x05,
   0x40, 0xa0, 0x46, 0x0c, 0x42, 0xee, 0xe0, 0x40, 0x4a, 0x44, 0x6e, 0xc6,
   0xc4, 0x44, 0xe2, 0x48, 0xea, 0x26, 0x6a, 0xa6, 0x4e, 0x68, 0xe2, 0xaa,
   0xa6, 0xe6, 0xa4, 0xea, 0x44, 0x24, 0x44, 0x00, 0xc0, 0xc6, 0xec, 0xee,
   0x46, 0xa4, 0xe4, 0x46, 0xc6, 0xc6, 0xae, 0xaa, 0xaa, 0x66, 0xc4, 0x00,
   0x00, 0xe0, 0x2c, 0x0a, 0x42, 0x44, 0x00, 0x20, 0x4a, 0x82, 0x88, 0x4a,
   0x8a, 0x00, 0x04, 0x04, 0xa2, 0x2a, 0x2a, 0xa2, 0x4a, 0xaa, 0xa2, 0xaa,
   0xe2, 0x8a, 0xa4, 0xea, 0x4a, 0x22, 0x48, 0x00, 0xa0, 0x2a, 0x6a, 0xa4,
   0x4a, 0x64, 0xe4, 0xaa, 0xaa, 0x42, 0xa4, 0xea, 0xa4, 0x44, 0x44, 0x00,
   0x40, 0xa0, 0x86, 0x0e, 0x24, 0x0a, 0x06, 0x24, 0xe6, 0x6e, 0x68, 0x44,
   0x44, 0x64, 0xe8, 0x42, 0xae, 0xc6, 0xe6, 0xc2, 0xea, 0xa4, 0xae, 0x4a,
   0xc2, 0x6a, 0xe4, 0xa4, 0x4a, 0x6e, 0x68, 0x00, 0xe0, 0xc6, 0xcc, 0x84,
   0x4a, 0xa4, 0xa4, 0x4a, 0xc6, 0x62, 0xec, 0xe4, 0x4a, 0xcc, 0x64, 0x00,
   0x00, 0x00, 0x04, 0x00, 0x00, 0x00, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xf0, 0x00, 0x00, 0x00, 0xe0,
   0x00, 0x06, 0x00, 0x00, 0x82, 0x00, 0x00, 0x00, 0x20, 0x00, 0x00, 0x00 };

// unpacked font (2 bpp)
byte font[FONT_BYTES*2];
byte font_init_done = FALSE;

// internal functions
void init_screen(void);
byte double_bits(byte);
void init_font(void);

// init graphics to either single or double buffered graphics
void init_graphics(int screens, byte *buffer)
{
	uint a;  // address as integer
	byte s;				//  screen index
	byte i, j, k;      	// counters
	byte *dl;          //  display list pointer instruction pointer
	uint *p;		// pointer to 16 bits
	uint addr, diff, bound, last, end;	// 1K/4K boundary alignment variables
	// init font once
	if (!font_init_done)
	{
		init_font();
		font_init_done = TRUE;
	}
	addr = (uint) buffer;     // addr = integer value of the address of top of buffer
	for (s = 0; s < screens; s++)  // for each screen (foreground/background)
	{
		bound = ((addr >> 10) + 1) << 10; // address of 1K boundary from addr address
		diff = bound - addr;		       // number of bytes from addr to boundary
		if (diff < LEN)	  // not enough bytes before 1K crossing!
		addr = bound;		// set addr of display list to beginning of next 1K...
		end = addr + LEN;		// end of display list
		display[s] = (byte*) addr;  // set the display list pointer
		addr = end;			// point addr at end of display list
		bound = ((addr >> 12) + 1) << 12 ;	// first 4K boundary crossing from addr
		diff = bound - addr;			// number of bytes from addr to 4K boundary
		if (diff >= 90*SCANLINE) // we need at least 90 scanlines 
		{
			last = diff / SCANLINE;          // save the # of scanlines on first 4K page
			addr = bound - last * SCANLINE;  // set the scanline/4K aligned screen address
		}
		else  // if we don't have at least 90 lines, then the first 4K is unusable
		{
			last = 102;		// omit the first 4K and alloc next 4K with 102 (max) scanlines
			// set screen to boundary plus remainder bytes (4K-102 scanlines .. 4096-102*40)
			addr = bound + 16;  
		}
		screen[s] = (byte*) addr;         // set screen pointer
		end = addr + SIZE;        // compute end of screen
		// fill scanline table now
		a = addr;         // set address to begining of screen
		for (i = 0; i < HEIGHT; i++)
		{
			scan_high[s][i] = a >> 8;  // set hi/lo byte scanline tables
			scan_low[s][i] = a & 0xff; 
			a += SCANLINE;      // next scanline
		}
		// create display list using the screen address we have just aligned to 4K boundary
		dl = display[s];			// use byte pointer dl for creating display list
		// emit three blank text lines (24 scanlines)
		for (i = 0; i < 3; i++)
		dl[i] = 0x70;
		// emit first LMS instruction in mode E
		dl[i++] = (1<<6) | MODE;	// emit mode E instruction and set bit six (LMS)
		p = (uint*) &dl[i];				// point at LMS data (address of scanline)
		*p = addr;		// set LMS data (address of screen)
		i += 2;				// LMS data took two bytes
		// emit first chunk of scanlines before 4K boundary
		k = last - 1;			// number of scanlines on first 4K minus 1 (LMS)
		for (j = 0; j < k; j++)
		dl[i++] = MODE;		// mode E scanline
		// emit second LMS instruction in mode E
		dl[i++] = (1<<6) | MODE;	// emit mode E instruction and set bit six (LMS)
		p = (uint*) &dl[i];				// point at LMS data (address of scanline)
		*p = addr + last * SCANLINE;		// set LMS data (screen 4K boundary)
		i += 2;				// LMS data took two bytes
		// emit second chunk of scanlines after 4K boundary
		k = (HEIGHT - last) - 1;	// number of scanlines remaining on second 4K minus 1(LMS)
		for (j = 0; j < k; j++)
		dl[i++] = MODE;		// mode E scanline 
		// finally emit jump and wait for vblank instruction (JVB)
		dl[i++] = 0x41;		// emit opcode
		p = (uint*) &dl[i];				// point at instruction data
		*p = (uint) display[s];		// set data (address of the beginning of this display list)
		// next screen
		addr = end;
	}
	buffer_usage = (long) addr - (long) buffer;
	// save color palette
	color_val[0] = COLOR[4];
	color_val[1] = COLOR[0];
	color_val[2] = COLOR[1];
	color_val[3] = COLOR[2];
	// save OS ANTIC registers
	sdmctl_val = *SDMCTL;
	sdlstl_val = *SDLSTL;
	// set zero page pixel mask table address for drawing functions
	addr = (uint) pixel_mask;
	*PM = addr;
	// init screen selector
	current_screen = 0;
	// install screen switcher for double buffered graphics
	if (screens > 1)
	{
		// init screen switcher vertical blank interrupt service routine
		request_swap = FALSE;  // init signal
		set_vbv((uint) screen_swapper);  // install the vertical blank routine
	}
	set_color(0);				// set bg color
	// clear buffer(s)
	for (i = 0; i < screens; i++)
	{
		if (screens > 1)  // if double buffered graphics
		swap_screen();				// switch screen in double buffered graphics context
		else			// init single buffered graphics
		{
			// set single buffer display list
			*SDMCTL = 0;
			*SDLSTL = (uint) display[0];
			*SDMCTL = sdmctl_val;
			// init scanline pointers
			init_screen();
		}
		clear_screen();					// clear this screen
	}
	// save number of screens being used
	screen_count = screens;
}

long graphics_memory_usage(void)
{
	return buffer_usage;
}

uint __fastcall__ set_vbv(uint addr)
{
	// because this is a fastcall function, the high byte of addr is already in the x register
	__asm__("tay");		// therefore transfer low byte of addr to y register (it was in the accumulator)
	__asm__("lda #6");      // set interrupt service selector (6 == immediate)
	__asm__("jsr $e45c"); 	// call the OS set vertical blank interrupt
	return addr;            // return addr so compiler doesn't complain addr wasn't used
}

void exit_graphics(void)
{
	// if double buffered graphics were used
	if (screen_count > 1)
	{
		// restore the ATARI OS vertical blank routine vector
		set_vbv(0xe45f);
	}
	// restore saved color palette
	set_palette(color_val);
	// turn off ANTIC
	*SDMCTL = 0;
	// load saved display list
	*SDLSTL = sdlstl_val;
	// turn ANTIC back on
	*SDMCTL = sdmctl_val;
	//printf("  freed %d screens\n", screen_count);
}

// call this function to switch which screen is displayed as well as which screen to draw to
void swap_screen(void)
{
	// do the time critical display list swap in the vbi routine
	request_swap = TRUE;		// request screen swap
	while (request_swap);		// wait for completion by screen swapper
	// switch screen selector now to draw to the background buffer
	current_screen ^= 1;
	// init background buffer scanline pointers
	init_screen();
}

// call this function to switch which screen to draw to only
void  toggle_screen(void)
{
	// switch screen selector now to draw to
	current_screen ^= 1;
	// init background buffer scanline pointers
	init_screen();
}

// initialize current screen (don't call this directly.. internal function!)
void init_screen(void)
{
	// set scanline table pointer for drawing routines
	*SH = (uint) &scan_high[current_screen][0];
	*SL = (uint) &scan_low[current_screen][0];
}

// this is the time critical display list switcher vertical blank interrupt (vbi) routine called when TV vblank occurs
void screen_swapper(void)              // (don't call this directly.. interrupt routine!)
{
	// this will be called every vertical blank, but does nothing until screen switch is requested
	if (request_swap)
	{
		// load the new display list
		*SDLSTL = (uint) display[current_screen];
		// signal completion
		request_swap = FALSE;
	}
	// pass control back to the ATARI OS vertical blank interrupt handler
	__asm__("jmp $e45f");
}

// draw filled polygon specified by n x,y pairs contained in v array 
void fill_polygon(byte *v, byte n)
{
	byte i, j;  // vertex indexes
	byte k, l;   // scaled vertex indexes
	byte y;  // y coord value
	byte y_min, y_max;    // y-span of polygon
	y_max = 0;         // init y  min/max to max/min values
	y_min = HEIGHT-1;
	for (i = 0; i < n; i++)
	{
		// get next vertex index
		j = i + 1;
		if (j == n)  // wrap around to first vertex at the end 
		j = 0;
		// scale indexes by 2
		k = i << 1;
		l = j << 1;
		// get y value of current vertex
		y = v[k+1];
		// test for new min/max y values
		if (y < y_min)  // set new min/max if found
		y_min = y;
		if (y > y_max)
		y_max = y;
		// set the polygon edge data now
		polygon_set_edge(v[k], y, v[l], v[l+1]);
	}
	// fill the scanlines with edges that were set
	polygon_fill_edge(y_min, y_max);
}

// draw polygon specified by n x,y pairs contained in v array 
void draw_polygon(byte *v, byte n)
{
	byte i, j;  // vertex indexes
	byte k, l;   // scaled vertex indexes
	for (i = 0; i < n; i++)
	{
		// get next vertex index
		j = i + 1;
		if (j == n)  // wrap around to first vertex at the end 
		j = 0;
		// scale indexes by 2
		k = i << 1;
		l = j << 1;
		draw_line(v[k], v[k+1], v[l], v[l+1]);
	}
}


void __fastcall__ set_color(byte c)
{
	*CL = pixel_fill[c];
}

void __fastcall__ set_palette(byte *p)
{
	COLOR[4] = p[0];
	COLOR[0] = p[1];
	COLOR[1] = p[2];
	COLOR[2] = p[3];
}

// unpack the 1-bit 4x6 font to 2 bits per pixel
void init_font(void)
{
	byte a, b, c;				// byte nibbles a & b, whole byte c
	int i, j;							// index into one bit font i, index into two bit font j
	j = 0;
	for (i = 0; i < FONT_BYTES; i++)
	{
		c = font4x6[i];								// get a byte out of 1 bit font
		b = c & 0xf0;									// separate two nibbles in font byte
		a = (c & 0x0f) << 4;						// make both upper nibbles
		font[j++] = double_bits(a);			// output two bytes to result
		font[j++] = double_bits(b);			// note the nibbles were reversed (XBM format)
	}
}

byte double_bits(byte n)				// double each bit in upper nibble
{
	byte i, b;				// index, bit
	byte bits;				// result
	bits = 0;
	for (i = 0; i < 4; i++)			// for each bit in source
	{
		bits >>= 2;						// shift result by two bits
		b = n & 0x80;				// get highest bit in source
		if (b)								// if source bit was set
		bits |= 0xc0;					// set two bits in result
		n <<= 1;							// shift source by one bit
	}
	return bits;
}

