
/*

	graphics.h
	
		8-bit ATARI computer graphics library functions and variables
		
	author:
	
		Bart Jaszcz
	
*/

#include "global.h"

// assembler interface zero page variables
#define SH ((uint*) 0xD4)				// scanline table pointer (high bytes)
#define SL ((uint*) 0xD6)				// scanline table pointer (low bytes)
#define PM ((uint*) 0xD8)				// pixel mask table pointer
#define X0 ((byte*) 0xDA)				// 1st x coordinate	
#define Y0 ((byte*) 0xDB)				// 1st y coordinate
#define X1 ((byte*) 0xDC)				// 2nd x coordinate
#define Y1 ((byte*) 0xDD)				// 2nd y coordinate
#define CL ((byte*) 0xDE)				// byte filled with current drawing color
#define GP ((byte*) 0xDF)				// general purpose byte
#define SC  ((uint*) 0xE0)				// 16 scanline pointers ($E0-$FF)

// ANTIC mode E display data
#define MODE 0xe		         		// graphics mode
#define BPP 2		              			// bits per pixel
#define WIDTH 160		        	  	// width of screen 
#define HEIGHT 192 					// height of screen
#define SIZE 7680					// bytes per screen
#define SCANLINE 40 					// bytes per scanline
#define LEN 202						// display list length
#define BUF (SIZE+LEN)					// memory per screen buffer

// OS shadow variables for ANTIC registers:
#define SDMCTL ((byte*) 0x22f)		// control register
#define SDLSTL ((uint*) 0x230)		// display list pointer 
#define COLOR ((byte*) 0x2c4)		// color palette registers

// ROM font location
#define FONT ((byte*) 0xe000)

// assembler functions
extern void xy_plot_0(void);
extern void xy_plot_1(void);
extern void h_line(void);
extern void v_line(void);
extern void line(void);
extern void set_edge(void);
extern void fill_edge(void);
extern void fill_rect(void);
extern void print_text(void);

// C functions
void init_graphics(int, byte*);
void draw_polygon(byte*, byte);
void fill_polygon(byte*, byte);
void swap_screen(void);
void toggle_screen(void);
void __fastcall__ set_color(byte);
void __fastcall__ set_palette(byte*);
long graphics_memory_usage(void);
void exit_graphics(void);

// special functions
uint __fastcall__ set_vbv(uint);		// vblank interrupt install routine
void screen_swapper(void);           		// vblank interrupt handler routine

// high byte of an 16 bit integer as a CC65 macro
#define high_byte(x) 		\
		(__AX__ = (x),		\
		asm ("txa"),			\
		asm ("ldx #$00"),	\
		__AX__)

// C to assembler interface
#define horizontal_line(x, x1, y) \
			(*X0 = x, *X1 = x1, *Y0 = y, h_line())
#define vertical_line(x, y, y1) \
			(*X0 = x, *Y0 = y, *Y1 = y1, v_line())
#define draw_line(x, y, x1, y1) \
			(*X0 = x, *Y0 = y, *X1 = x1, *Y1 = y1, line())
#define polygon_set_edge(x, y, x1, y1) \
			(*X0 = x, *Y0 = y, *X1 = x1, *Y1 = y1, set_edge())
#define polygon_fill_edge(y, y1) \
			(*Y0 = y, *Y1 = y1, fill_edge())
#define fill_rectangle(x, y, x1, y1) \
			(*X0 = x, *Y0 = y, *X1 = x1, *Y1 = y1, fill_rect())
#define draw_rectangle(x, y, x1, y1) \
			(*X0 = x, *X1 = x1, *Y0 = y1, h_line(), *Y0 = y, h_line(), *Y1 = y1, v_line(), *X0 = x1, v_line())
#define clear_screen() \
			(fill_rectangle(0,0,WIDTH-1,HEIGHT-1))
#define plot(x, y) \
			(*X0 = x, *Y0 = y, xy_plot_0())
#define print_text(col, row, str) \
			(*X0 = col, *Y0 = row, *X1 = (byte) str, *Y1 = high_byte((unsigned int) str), print_text())



