;
; set_edge.s						9/27/2008
;
;   Polygon set edge routine.  Sets the appropriate edge bounds (left or right) of
; a polygon which will be filled by the fill_edge routine.
;
; author:
;
;   Bart Jaszcz
;

	
		.include "math16.inc"
		.include "graphics.inc"
		.importzp ptr1, ptr2, ptr3, ptr4, tmp3, tmp4
		.import _left_edge, _right_edge
		.export _set_edge

; 16 bit variables:

edge	= ptr1						; edge array pointer
f		= ptr2						; Bresenham's variable
dx		= ptr3						; delta x
dy		= ptr4						; delta y

; 8 bit variables

sx = tmp3							; step direction flags
sy = tmp4
	
.proc _set_edge

_set_edge:		lda Y1				; check for horizontal edges
				cmp Y0
				bne cont			; continue if diff not 0
				rts					; otherwise return
				
cont:			lda X1				; get X1 coordinate
				sta dx				; store in low byte of dx
				lda #0				; clear high byte of dx
				sta dx+1
				
				lda X0				; get X0 coordinate
				sta f				; store in low byte of f
				lda #0				; clear high byte of f
				sta f+1

				sub16 dx, f			; dx -= f (X1 -= X0)
				lda dx+1			; load high byte of dx
				bmi @negate			; if dx is negative, negate it
				lda #1				; otherwise dx is positive
				sta sx				; set positive step-x flag
				bpl calc_dy			; continue to calc dy code

@negate:		neg16 dx			; negate dx so it's now positive
				lda #0				; set step-x flag to 0 (negative step)
				sta sx

calc_dy:		lda Y1				; get Y1 coordinate
				sta dy				; store in low byte of dy
				lda #0				; clear high byte of dy
				sta dy+1
				lda Y0				; get Y0 coordinate
				sta f				; store in low byte of f
				lda #0				; clear high byte of f
				sta f+1

				sub16 dy, f			; dy -= f (Y1 -= Y0)
				lda dy+1			; load high byte of dy
				bmi @negate			; if dy is negative, negate it
				lda #1				; otherwise dy is positive
				sta sy				; set positive step-y flag
				lda #<_left_edge		; get low byte of left_edge address
				sta edge			; set low byte of edge pointer
				lda #>_left_edge		; get high byte of left_edge address
				sta edge+1			; set high byte of edge pointer
				bpl init			; continue to init code

@negate:		neg16 dy			; negate dx so it's now positive
				lda #0				; set step-x flag to 0 (negative step)
				sta sy
				lda #<_right_edge		; get low byte of left_edge address
				sta edge			; set low byte of edge pointer
				lda #>_right_edge		; get high byte of left_edge address
				sta edge+1			; set high byte of edge pointer

init:			ldy Y0				; output endpoint 0
				lda X0
				sta (edge), y
				ldy Y1				; output endpoint 1
				lda X1
				sta (edge), y

				lda dx				; compare dx and dy
				cmp dy
				bcc y_init			; if dx < dy jump to y init

x_init:			lsr a				; divide dx by 2 (interation count)
				tax					; store in x register (loop counter)
				asl16 dy			; dy <<= 1
				dif16 f, dy, dx		; f = dy - dx
				asl16 dx			; dx <<= 1
x_loop:			lda f+1				; load high byte of f into a
				bmi change_x		; if f < 0 jump to change x
				lda sy				; load step-y flag
				beq @skip			; 0 means negative, jump over positive code
				dec Y1				; decrement Y1 coordinate				
				inc Y0				; increment Y0 coordinate
				bne subtract_dx		; relative jump to subtract dx (always taken)
@skip:			inc Y1				; increment Y1
				dec Y0				; decrement Y0
subtract_dx:	sub16 f, dx			; f -= dx
change_x:		lda sx				; load step-x flag
				beq @skip			; 0 means negative, jump over positve code
				dec X1				; decrement X1 coordinate
				inc X0				; increment X0 coordinate
				bne x_out			; branch to x out (always taken)
@skip:			inc X1				; increment X1 coordinate
				dec X0				; decrement X0 coordinate
x_out:			ldy Y0				; output endpoint 0
				lda X0
				sta (edge), y
				ldy Y1				; output endpoint 1
				lda X1
				sta (edge), y
				add16 f, dy			; f += dy
				dex					; decrement loop counter (x register)
				bne x_loop			; jump to top of x loop
				rts					; return from set_edge routine

y_init:			lda dy				; load y difference
				lsr a				; divide dy by 2 (interation count)
				tax					; store in x register (loop counter)
				asl16 dx			; dy <<= 1
				dif16 f, dx, dy		; f = dx - dy
				asl16 dy			; dy <<= 1
y_loop:			lda f+1				; load high byte of f into a
				bmi change_y		; if f < 0 jump to change y
				lda sx				; load step-x flag
				beq @skip			; 0 means negative, jump over positive code
				dec X1				; decrement Y1 coordinate				
				inc X0				; increment Y0 coordinate
				bne subtract_dy		; relative jump to subtract dy (always taken)
@skip:			inc X1				; increment Y1
				dec X0				; decrement Y0
subtract_dy:	sub16 f, dy			; f -= dy
change_y:		lda sy				; load step-y flag
				beq @skip			; 0 means negative, jump over positve code
				dec Y1				; decrement Y1 coordinate
				inc Y0				; increment Y0 coordinate
				bne y_out			; branch to y out (always taken)
@skip:			inc Y1				; increment Y1 coordinate
				dec Y0				; decrement Y0 coordinate
y_out:			ldy Y0				; output endpoint 0
				lda X0
				sta (edge), y
				ldy Y1				; output endpoint 1
				lda X1
				sta (edge), y
				add16 f, dx			; f += dx
				dex					; decrement loop counter (x register)
				bne y_loop			; jump to top of y loop
				rts					; return from set_edge routine

.endproc




